{******************************************************************************
 This Datasource component gives you access to datasets defined on another form.
 It works with Delphi 1 and Delphi 2.

 This component with source code is provided as freeware.
 You may use it in your programs or distribute it as you wish.

 To install, copy the files DATASRC.PAS and DATASRC.DCR into your component
 directory. Then from the Options menu of Delphi, select Install Components.
 Hit the Add button and type the full directory and file name of this
 component, or browse it.

 To use, place a TDatahost component on a form that has some datasets defined
 on it. Name it. All datasets defined on that form can be visible to any
 TRemoteDatasource in your project. Now place a TRemoteDatasource on another
 form. Click the datahost property and choose the datahost you just defined.
 Then click the dataset property, and a list of all the datasets you defined
 on the other form will appear.

 NOTE:
   Datahosts needs unique names.

 If you run into problems or have questions,
 email gzimmer@usa.net

****************************************************************************** }

unit DataSrc;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, DsgnIntf, Typinfo;

procedure Register;

type TDataHost = class (TComponent)
  private
  protected
    procedure SetName(const NewName: TComponentName); override;
  public
    MyRemotes: TList;
    procedure Linkup;
    procedure Loaded; override;
    constructor Create(aowner: tcomponent); override;
    destructor Destroy;override;
  end;

type TRemoteDatasource = class (TDataSource)
  private
    FHost: TDataHost;
    FHostName: string;
    FDatasetName: string;
    function FindHost(const Hostname: string):TDataHost;
  protected
    procedure LinkHost(Host: TDataHost);
    procedure Readdata(reader:treader);
    procedure Writedata(writer:twriter);
    procedure SetDataset(value:tdataset);
    function GetDataset: tdataset;
    function GetDataHost: tdatahost;
    procedure SetDataHost(value: tdatahost);
    procedure DefineProperties(filer: tfiler);override;
  public
    constructor Create(aowner: tcomponent);override;
    destructor Destroy; override;
    procedure Loaded; override;
  published
    property Datahost: TDatahost read getdatahost write setdatahost stored false;
    property Dataset: tdataset read getdataset write setdataset stored false;
  end;

implementation

var HostList: TList;
    RemoteList: TList;

{------------------}
{TDatahost}
{------------------}
constructor TDataHost.Create(aowner: tcomponent);
begin
  inherited Create(aowner);
  MyRemotes:=TList.Create; {list of remotes linked to this host}
  HostList.add(self);
end;

destructor TDataHost.Destroy;
begin
  {unlink all MyRemotes attached to this Host}
  while MyRemotes.count>0 do begin
    TRemoteDatasource(MyRemotes[0]).Linkhost(nil); {linkhost(nil) decreases count because of .pack}
  end;
  MyRemotes.free;
  MyRemotes:=nil;

  HostList.remove(self);
  HostList.pack;
  inherited Destroy;
end;

{find any RemoteDataSources waiting for link}
procedure TDataHost.Linkup;
var n: integer;
begin
  for n:=0 to RemoteList.count-1 do begin
    if comparetext(TRemoteDatasource(RemoteList[n]).FHostname,name)=0 then begin
       TRemoteDatasource(RemoteList[n]).DataHost:=self;
    end;
  end;
end;

function StandardNameNumber(const Name, ClassName: TComponentName): Integer;
begin
  if (Length(Name) > Length(ClassName)) and
     (Copy(Name, 1, Length(ClassName)) = ClassName) then
  begin
    try
      result := StrToInt(Copy(Name, Length(ClassName) + 1, 255));
    except
      result := -1;
    end;
  end
  else
    if (Name = ClassName) then
      result := 0
    else
      result := -1;
end;

function HostNameExists(const Name: TComponentName): Boolean;
var
  i: Integer;
begin
  result := true;
  for i := 0 to HostList.count -1 do
    if (TDatahost(HostList[i]).Name = Name) then
      exit;
  result := false;
end;

procedure TDatahost.SetName(const NewName: TComponentName);
const
  MsgUniqueNamesRequired = 'Unique names for all datahosts required';
  Impossible = 100;
var
  NewNameChecked: TComponentName;
  oldname: TComponentName;
  StandardComponentName: TComponentName;
  n: integer;
  NameNumber: Integer;
function TryToSetNewName: Boolean;
begin
  try { If name exists, SetName raises an EComponentError exception: }
    inherited SetName(NewNameChecked);
    result := true;
  except
    result := false;
  end;
end;
begin { SetName }
  oldname := Name;
  NewNameChecked := NewName;
  if (oldname <> NewName) and IsValidIdent(NewName) then
  begin
  { a TDatahost named with NewName exists on another form already: }
    if HostNameExists(NewNameChecked) then
    begin
      if (csDesigning in ComponentState) and (csLoading in ComponentState) then
      begin
      { The Owner-form was closed while the name of another TDatahost changed.
        We cannot rename it here, because the name in the DFN-file has to be
        changed too. }
        inherited SetName(NewNameChecked); { Set name anyway, but warn: }
        MessageDlg(MsgUniqueNamesRequired + '.' + #10 +
                   'Please rename ' + NewNameChecked + '.',
                   mtWarning, [mbOk], 0);
      end
      else
      begin
        n := 0;
        StandardComponentName := ClassName; { 'TDataHost' }
        if (Upcase(StandardComponentName[1]) = 'T') then
          System.Delete(StandardComponentName, 1, 1); { 'DataHost' }
        repeat
          NameNumber := StandardNameNumber(
            NewNameChecked, StandardComponentName);
          if (NameNumber < 0) then { user-defined name: }
            raise EComponentError.Create(MsgUniqueNamesRequired); { -> exit }
        { DataHostX exists, try DataHostX+1: }
          NewNameChecked := StandardComponentName + IntToStr(NameNumber + 1);
          Inc(n);
        until ((not HostNameExists(NewNameChecked)) and TryToSetNewName) or
              (n = Impossible);
        if (NameNumber = Impossible) then
        { never happens, but I don't want to risk an endless loop in design-mode }
          raise EComponentError.Create('Unknown error at TDatahost.SetName');
      end;
    end
    else
      inherited SetName(NewNameChecked);
  end
  else
    inherited SetName(NewNameChecked);
{ for debug only:
  if (NewNameChecked <> NewName) then
  begin
    MessageDlg(MsgUniqueNamesRequired + #10 +
               NewName + ' was renamed to ' + NewNameChecked,
               mtInformation, [mbOk], 0);
  end;
{}
  if oldname<>NewNameChecked then begin {update links if name changed}
    for n:=0 to RemoteList.count-1 do begin
      if comparetext(TRemoteDatasource(RemoteList[n]).FHostname,oldname)=0 then
      begin
         TRemoteDatasource(RemoteList[n]).Datahost:=self;
      end;
    end;
  end;
end;

procedure TDataHost.Loaded;
begin
  inherited loaded;
  Linkup;
end;

{------------------}
{TRemoteDatasource}
{------------------}

constructor TRemoteDatasource.Create(aowner: tcomponent);
begin
  if aowner=nil then raise exception.Create('Owner required');
  inherited Create(aowner);
  RemoteList.add(self);
end;

destructor TRemoteDatasource.Destroy;
begin
  linkhost(nil); {unlink from host}
  RemoteList.remove(self);
  RemoteList.pack;
  inherited Destroy;
end;

procedure TRemoteDatasource.Loaded;
begin
  inherited loaded;
  linkhost(findhost(FHostname));
end;

function TRemoteDatasource.getDataset: tdataset;
begin
  result:=inherited dataset;
end;

procedure TRemoteDatasource.setDataset(value: tdataset);
begin
  inherited dataset:=value;
  if value<>nil then
     FDatasetname:=value.name; {save for later}
end;

procedure TRemoteDatasource.SetDataHost(value: tdatahost);
begin
   if value=nil then begin
     linkhost(nil);
     FHostname:='';
   end else begin
     linkhost(value);
     Fhostname:=value.name;
   end;
end;

function TRemoteDatasource.GetDataHost: tdatahost;
begin
  result:=findhost(fhostname);
end;

procedure TRemoteDatasource.LinkHost(Host: TDataHost);
begin
  {if already linked to some Host, unlink from it}
  if FHost<>nil then begin
    FHost.MyRemotes.remove(self);
    FHost.MyRemotes.pack; {prevent nil in list}
    FHost:=nil;
  end;

  {link to new host}
  if Host<>nil then Begin
    {make sure they didnt cast something as tdatahost... just a precaution}
    if not (host is TDatahost) then
      raise exception.Create('Invalid DataHost');

    Host.MyRemotes.add(self);
    dataset:=tdataset(host.owner.findcomponent(fdatasetname));
    Fhost:=Host;
  end;
end;

{function to find a Host in the Host list}
function TRemoteDatasource.findHost(const Hostname:string):TDataHost;
var n: integer;
begin
  result:=nil;

  for n:=0 to HostList.count-1 do begin
    if comparetext(TDataHost(HostList[n]).name,Hostname)=0 then begin
       result:=HostList[n];
       break;
    end;
  end;
end;

procedure TRemoteDatasource.Readdata(reader:treader);
begin
   reader.readlistbegin;
   FDataSetName:=reader.readstring;
   FHostName:=reader.readstring;
   reader.readlistend;
end;

procedure TRemoteDatasource.Writedata(writer:twriter);
begin
   writer.writelistbegin;
   writer.writestring(FDataSetName);
   writer.writestring(FHostName);
   writer.writelistend;
end;

procedure TRemoteDatasource.DefineProperties(filer: tfiler);
begin
   inherited defineproperties(filer);
   filer.DefineProperty('RemoteData',Readdata,WriteData,(FDatasetName<>'') or (FHostName<>''));
end;

{------------------}
{ these property editors will be removed from the final .EXE by the linker}
{------------------}

{------------------}
{ TRemoteDataSetProperty }
{------------------}
type
  TRemoteDataSetProperty = class(TStringProperty)
  public
    procedure GetValues(Proc: TGetStrProc); override;
    function GetAttributes: TPropertyAttributes; override;
    procedure SetValue(const Value: string); override;
    function GetValue:string; override;
  end;

function TRemoteDatasetProperty.GetValue: string;
var component:tcomponent;
begin
   component:=Tcomponent(getordvalue);
   if component=nil then begin
      result:=TRemoteDatasource(getcomponent(0)).FDatasetname;
   end else begin
      result:=component.name;
   end;
end;

procedure TRemoteDatasetProperty.SetValue(const Value: string);
var
  Component: TComponent;
  source: TRemoteDatasource;
  host: TDatahost;
begin
  Component := nil;
  if Value <> '' then begin
     source:=TRemoteDatasource(getcomponent(0));
     if source<>nil then begin
       host:=source.findhost(source.fhostname);
       if host<>nil then begin
         if host.owner<>nil then begin
           Component := host.owner.FindComponent(Value);
           if (component=nil) or not (Component is GetTypeData(GetPropType)^.ClassType) then
             raise EPropertyError.Create('Invalid Property Value');
         end;
       end;
     end;
  end;
  SetOrdValue(longint(Component));
end;

function TRemoteDatasetProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList];
end;

procedure TRemoteDatasetProperty.GetValues(Proc: TGetStrProc);
var
  Linked: Boolean;
  I,J: Integer;
  Component: TComponent;
  source: TRemoteDatasource;
  host: TDatahost;
  Form: tform;
begin
  source:=TRemoteDatasource(getcomponent(0));
  if source<>nil then begin
    host:=source.findhost(source.fhostname);
    if host=nil then begin
       raise EPropertyError.Create(
         'Could not find Datahost <' + source.fHostname + '>.' + #10 +
         'Please open the form where it is defined in first');
    end;
    Form:=tform(host.owner);
    for I := 0 to Form.ComponentCount - 1 do
    begin
      Component := Form.Components[I];
      if (Component is TDataSet) and (Component.Name <> '') then
      begin
        Linked := False;
        for J := 0 to PropCount - 1 do
          Linked := Linked or
            TDataSource(GetComponent(J)).IsLinkedTo(TDataSet(Component));
        if not Linked then Proc(Component.Name);
      end;
    end;
  end;
end;


type
  TDataHostProperty = class(TComponentProperty)
  public
    procedure GetValues(Proc: TGetStrProc); override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

function TDataHostProperty.GetValue: string;
var component:tcomponent;
begin
   component:=Tcomponent(getordvalue);
   if component=nil then begin
      result:=TRemoteDatasource(getcomponent(0)).FHostname;
   end else begin
      result:=component.name;
   end;
end;

procedure TDataHostProperty.GetValues(Proc: TGetStrProc);
var
  J: Integer;
begin
  for J := 0 to hostlist.count - 1 do begin
    Proc(tdatahost(hostlist[j]).name);
  end;
end;

procedure TDataHostProperty.SetValue(const value: string);
var host: TDatahost;
begin
  host:=TRemoteDatasource(getcomponent(0)).findhost(value);
  if (host=nil) and (value<>'') then begin
     messagedlg(value+' is not found.'#10'Open the form where it is defined and then try again.',mtInformation,[mbOk],0);
  end else begin {if host not found, dont clear it out, the Form
                 it lives on may not be opened yet!}
     SetordValue(longint (host));
  end;
end;

procedure Register;
begin
  RegisterComponents('Data Access', [TDataHost, TRemoteDatasource]);
  RegisterPropertyEditor(TypeInfo(TDataset), TRemoteDatasource, 'DataSet', TRemoteDatasetProperty);
  RegisterPropertyEditor(TypeInfo(TDatahost), TRemoteDatasource, 'Datahost', TDataHostProperty);
end;

begin
  HostList:=TList.Create;  {these will be freed when app exits}
  RemoteList:=TList.Create;
end.

